From e3c91efe9a759caffc6580bc5475684338170fbf Mon Sep 17 00:00:00 2001
From: Martell Malone <martellmalone@gmail.com>
Date: Wed, 24 Dec 2014 16:37:27 +0000
Subject: [PATCH] Create a dummy d3d11 vout plugin initially

---
 configure.ac                          |   7 +-
 modules/MODULES_LIST                  |   1 +
 modules/video_output/Makefile.am      |  12 +
 modules/video_output/msw/common.c     |  10 +-
 modules/video_output/msw/common.h     |  40 ++++
 modules/video_output/msw/direct3d11.c | 407 ++++++++++++++++++++++++++++++++++
 modules/video_output/msw/events.c     |   2 +-
 7 files changed, 472 insertions(+), 7 deletions(-)
 create mode 100644 modules/video_output/msw/direct3d11.c

diff --git a/configure.ac b/configure.ac
index e6760c6..f0ead5a 100644
--- a/configure.ac
+++ b/configure.ac
@@ -3254,7 +3254,12 @@ AS_IF([test "${enable_directx}" != "no"], [
     #include <GL/gl.h>
   ])
 
-  dnl Direct3D
+  dnl Direct3D11
+  AC_CHECK_HEADERS(d3d11.h, [
+    VLC_ADD_PLUGIN([direct3d11])
+  ])
+
+  dnl Direct3D9
   AC_CHECK_HEADERS(d3d9.h, [
     VLC_ADD_PLUGIN([direct3d9])
   ])
diff --git a/modules/MODULES_LIST b/modules/MODULES_LIST
index 0f5adc9..85023c4 100644
--- a/modules/MODULES_LIST
+++ b/modules/MODULES_LIST
@@ -95,6 +95,7 @@ $Id$
  * diracsys: BBC Dirac demuxer
  * direct2d: video output module using the Direct2D API
  * direct3d9: video output module using the Direct3D9 API
+ * direct3d11: video output module using the Direct3D11 API
  * directdraw: video output module using the DirectDraw API
  * directfb: Direct Framebuffer video output
  * directsound: audio output module using the DirectSound API
diff --git a/modules/video_output/Makefile.am b/modules/video_output/Makefile.am
index beb3644..fda9ed0 100644
--- a/modules/video_output/Makefile.am
+++ b/modules/video_output/Makefile.am
@@ -181,6 +181,18 @@ libdirect3d9_plugin_la_LDFLAGS = $(AM_LDFLAGS) -rpath '$(voutdir)'
 vout_LTLIBRARIES += $(LTLIBdirect3d9)
 EXTRA_LTLIBRARIES += libdirect3d9_plugin.la
 
+libdirect3d11_plugin_la_SOURCES = video_output/msw/direct3d11.c \
+	video_output/msw/common.c video_output/msw/common.h \
+	video_output/msw/events.c video_output/msw/events.h \
+	video_output/msw/builtin_shaders.h \
+	video_output/msw/win32touch.c video_output/msw/win32touch.h
+libdirect3d11_plugin_la_CPPFLAGS = $(AM_CPPFLAGS) \
+	-DMODULE_NAME_IS_direct3d11
+libdirect3d11_plugin_la_LIBADD = -lgdi32 -lole32 -luuid
+libdirect3d11_plugin_la_LDFLAGS = $(AM_LDFLAGS) -rpath '$(voutdir)'
+vout_LTLIBRARIES += $(LTLIBdirect3d11)
+EXTRA_LTLIBRARIES += libdirect3d11_plugin.la
+
 libdirectdraw_plugin_la_SOURCES = video_output/msw/directx.c \
 	video_output/msw/common.c video_output/msw/common.h \
 	video_output/msw/events.c video_output/msw/events.h \
diff --git a/modules/video_output/msw/common.c b/modules/video_output/msw/common.c
index 092b9bb..cac6613 100644
--- a/modules/video_output/msw/common.c
+++ b/modules/video_output/msw/common.c
@@ -72,7 +72,7 @@ int CommonInit(vout_display_t *vd)
 
     event_cfg_t cfg;
     memset(&cfg, 0, sizeof(cfg));
-#ifdef MODULE_NAME_IS_direct3d9
+#if defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct3d11)
     cfg.use_desktop = sys->use_desktop;
 #endif
 #ifdef MODULE_NAME_IS_directdraw
@@ -348,7 +348,7 @@ void UpdateRects(vout_display_t *vd,
                      SWP_NOCOPYBITS|SWP_NOZORDER|SWP_ASYNCWINDOWPOS);
 
     /* Destination image position and dimensions */
-#if defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct2d)
+#if defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct3d11) || defined(MODULE_NAME_IS_direct2d)
     rect_dest.left   = 0;
     rect_dest.right  = place.width;
     rect_dest.top    = 0;
@@ -425,7 +425,7 @@ void UpdateRects(vout_display_t *vd,
     /* Apply overlay hardware constraints */
     if (sys->use_overlay)
         AlignRect(&rect_src_clipped, sys->i_align_src_boundary, sys->i_align_src_size);
-#elif defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct2d)
+#elif defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct3d11) || defined(MODULE_NAME_IS_direct2d)
     /* Needed at least with YUV content */
     rect_src_clipped.left &= ~1;
     rect_src_clipped.right &= ~1;
@@ -465,7 +465,7 @@ static int CommonControlSetFullscreen(vout_display_t *vd, bool is_fullscreen)
 {
     vout_display_sys_t *sys = vd->sys;
 
-#ifdef MODULE_NAME_IS_direct3d9
+#if defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct3d11)
     if (sys->use_desktop && is_fullscreen)
         return VLC_EGENERIC;
 #endif
@@ -591,7 +591,7 @@ int CommonControl(vout_display_t *vd, int query, va_list args)
     case VOUT_DISPLAY_CHANGE_WINDOW_STATE: {       /* unsigned state */
         const unsigned state = va_arg(args, unsigned);
         const bool is_on_top = (state & VOUT_WINDOW_STATE_ABOVE) != 0;
-#ifdef MODULE_NAME_IS_direct3d9
+#if defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct3d11)
         if (sys->use_desktop && is_on_top)
             return VLC_EGENERIC;
 #endif
diff --git a/modules/video_output/msw/common.h b/modules/video_output/msw/common.h
index 5cb8c45..bc02898 100644
--- a/modules/video_output/msw/common.h
+++ b/modules/video_output/msw/common.h
@@ -25,6 +25,11 @@
 #ifdef MODULE_NAME_IS_directdraw
 # include <ddraw.h>
 #endif
+#ifdef MODULE_NAME_IS_direct3d11
+// TODO : look at dxgi1_2.h for 3D stuff
+# include <d3d11.h>
+# include <dxgi.h>
+#endif
 #ifdef MODULE_NAME_IS_direct3d9
 # include <d3d9.h>
 # include <d3dx9effect.h>
@@ -135,6 +140,41 @@ struct vout_display_sys_t
     ID2D1Bitmap            *d2_bitmap;                            /* D2 bitmap */
 #endif
 
+#ifdef MODULE_NAME_IS_direct3d11
+    bool allow_hw_yuv;    /* Should we use hardware YUV->RGB conversions */
+    /* show video on desktop window ? */
+    bool use_desktop;
+    struct {
+        bool is_fullscreen;
+        bool is_on_top;
+        RECT win;
+    } desktop_save;
+    vout_display_cfg_t cfg_saved; /* configuration used before going into desktop mode */
+
+    // core objects
+    HINSTANCE               hd3d11_dll;       /* handle of the opened d3d9 dll */
+    HINSTANCE               hd3d11x_dll;      /* handle of the opened d3d9x dll */
+
+    // TODO : a dual implementation with IDXGIFactory2 for 3D stuff
+    IDXGIFactory            d3dfactory;
+
+    int                     d3dregion_count;
+    struct d3d_region_t     *d3dregion;
+
+    picture_sys_t           *picsys;
+
+    /* */
+    bool                    reset_device;
+    bool                    reopen_device;
+    bool                    lost_not_ready;
+    bool                    clear_scene;
+
+    /* It protects the following variables */
+    vlc_mutex_t    lock;
+    bool           ch_desktop;
+    bool           desktop_requested;
+#endif
+
 #ifdef MODULE_NAME_IS_direct3d9
     bool allow_hw_yuv;    /* Should we use hardware YUV->RGB conversions */
     /* show video on desktop window ? */
diff --git a/modules/video_output/msw/direct3d11.c b/modules/video_output/msw/direct3d11.c
new file mode 100644
index 0000000..0688521
--- /dev/null
+++ b/modules/video_output/msw/direct3d11.c
@@ -0,0 +1,407 @@
+/*****************************************************************************
+ * direct3d11.c: Windows Direct3D11 video output module
+ *****************************************************************************
+ * Copyright (C) 2014 VLC authors and VideoLAN
+ *
+ * Authors: Martell Malone <martellmalone@gmail.com>
+ *
+ * This program is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
+ * (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+ * GNU Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program; if not, write to the Free Software Foundation,
+ * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
+ *****************************************************************************/
+
+#ifdef HAVE_CONFIG_H
+# include "config.h"
+#endif
+
+#include <vlc_common.h>
+#include <vlc_plugin.h>
+#include <vlc_vout_display.h>
+#include <vlc_charset.h>
+
+#include <windows.h>
+#include <d3d11.h>
+
+#include "common.h"
+#include "builtin_shaders.h"
+
+/*****************************************************************************
+ * Module descriptor
+ *****************************************************************************/
+static int  Open(vlc_object_t *);
+static void Close(vlc_object_t *);
+
+#define DESKTOP_LONGTEXT N_(\
+    "The desktop mode allows you to display the video on the desktop.")
+
+#define D3D11_HELP N_("Recommended video output for Windows 8 and later versions")
+
+vlc_module_begin ()
+    set_shortname("Direct3D11")
+    set_description(N_("Direct3D11 video output"))
+    set_help(D3D11_HELP)
+    set_category(CAT_VIDEO)
+    set_subcategory(SUBCAT_VIDEO_VOUT)
+
+    set_capability("vout display", 280)
+    add_shortcut("direct3d11")
+    set_callbacks(Open, Close)
+
+vlc_module_end ()
+
+/*****************************************************************************
+ * Local prototypes.
+ *****************************************************************************/
+static const vlc_fourcc_t d3d_subpicture_chromas[] = {
+    VLC_CODEC_RGBA,
+    0
+};
+
+static int  Open(vlc_object_t *);
+static void Close(vlc_object_t *object);
+
+static picture_pool_t *Pool  (vout_display_t *, unsigned);
+static void           Prepare(vout_display_t *, picture_t *, subpicture_t *subpicture);
+static void           Display(vout_display_t *, picture_t *, subpicture_t *subpicture);
+static int            Control(vout_display_t *, int, va_list);
+static void           Manage (vout_display_t *);
+
+static int  Direct3D11Create (vout_display_t *);
+static int  Direct3D11Reset  (vout_display_t *);
+static void Direct3D11Destroy(vout_display_t *);
+
+static int  Direct3D11Open (vout_display_t *, video_format_t *);
+static void Direct3D11Close(vout_display_t *);
+
+static int DesktopCallback(vlc_object_t *, char const *, vlc_value_t, vlc_value_t, void *);
+
+
+static int Open(vlc_object_t *object)
+{
+    vout_display_t *vd = (vout_display_t *)object;
+    vout_display_sys_t *sys;
+
+    OSVERSIONINFO winVer;
+    winVer.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
+    if((GetVersionEx(&winVer) && !object->b_force) 
+        && ((winVer.dwMajorVersion > 5 && winVer.dwMinorVersion < 1 ) 
+        || (winVer.dwMajorVersion < 6)))
+       // minor versions since Windows 7 are all above 0
+        return VLC_EGENERIC;
+
+    vd->sys = sys = calloc(1, sizeof(vout_display_sys_t));
+    if (!sys)
+        return VLC_ENOMEM;
+
+    if (Direct3D11Create(vd)) {
+        msg_Err(vd, "Direct3D11 could not be initialized");
+        Direct3D11Destroy(vd);
+        free(sys);
+        return VLC_EGENERIC;
+    }
+
+    sys->use_desktop = var_CreateGetBool(vd, "video-wallpaper");
+    sys->reset_device = false;
+    sys->reopen_device = false;
+    sys->lost_not_ready = false;
+    sys->allow_hw_yuv = var_CreateGetBool(vd, "directx-hw-yuv");
+    sys->desktop_save.is_fullscreen = vd->cfg->is_fullscreen;
+    sys->desktop_save.is_on_top     = false;
+    sys->desktop_save.win.left      = var_InheritInteger(vd, "video-x");
+    sys->desktop_save.win.right     = vd->cfg->display.width;
+    sys->desktop_save.win.top       = var_InheritInteger(vd, "video-y");
+    sys->desktop_save.win.bottom    = vd->cfg->display.height;
+
+    if (CommonInit(vd))
+        goto error;
+
+    /* */
+    video_format_t fmt;
+    if (Direct3D11Open(vd, &fmt)) {
+        msg_Err(vd, "Direct3D11 could not be opened");
+        goto error;
+    }
+
+    vout_display_info_t info = vd->info;
+    info.is_slow = true;
+    info.has_double_click = true;
+    info.has_hide_mouse = false;
+    info.has_pictures_invalid = true;
+    info.has_event_thread = true;
+    info.subpicture_chromas = NULL;
+
+    /* Interaction */
+    vlc_mutex_init(&sys->lock);
+    sys->ch_desktop = false;
+    sys->desktop_requested = sys->use_desktop;
+
+    vlc_value_t val;
+    val.psz_string = _("Desktop");
+    var_Change(vd, "video-wallpaper", VLC_VAR_SETTEXT, &val, NULL);
+    var_AddCallback(vd, "video-wallpaper", DesktopCallback, NULL);
+
+    video_format_Clean(&vd->fmt);
+    video_format_Copy(&vd->fmt, &fmt);
+    vd->info = info;
+
+    vd->pool    = Pool;
+    vd->prepare = Prepare;
+    vd->display = Display;
+    vd->control = Control;
+    vd->manage  = Manage;
+
+    if (sys->use_desktop && vd->cfg->is_fullscreen)
+        vout_display_SendEventFullscreen(vd, false);
+
+    // TODO : chage to VLC_SUCCESS when d3d11 is working
+    return VLC_EGENERIC;
+error:
+    Direct3D11Close(vd);
+    CommonClean(vd);
+    Direct3D11Destroy(vd);
+    free(vd->sys);
+    return VLC_EGENERIC;
+}
+
+static void Close(vlc_object_t *object)
+{
+    vout_display_t * vd = (vout_display_t *)object;
+
+    var_DelCallback(vd, "video-wallpaper", DesktopCallback, NULL);
+    vlc_mutex_destroy(&vd->sys->lock);
+
+    Direct3D11Close(vd);
+
+    CommonClean(vd);
+
+    Direct3D11Destroy(vd);
+
+    free(vd->sys);
+}
+
+static picture_pool_t *Pool(vout_display_t *vd, unsigned count)
+{
+    VLC_UNUSED(count);
+    return vd->sys->pool;
+}
+
+static void Prepare(vout_display_t *vd, picture_t *picture, subpicture_t *subpicture)
+{
+
+}
+
+static void Display(vout_display_t *vd, picture_t *picture, subpicture_t *subpicture)
+{
+    CommonDisplay(vd);
+}
+
+static int ControlResetDevice(vout_display_t *vd)
+{
+    return Direct3D11Reset(vd);
+}
+
+static int ControlReopenDevice(vout_display_t *vd)
+{
+    vout_display_sys_t *sys = vd->sys;
+
+    if (!sys->use_desktop) {
+        sys->desktop_save.is_fullscreen = vd->cfg->is_fullscreen;
+        sys->desktop_save.is_on_top     = sys->is_on_top;
+
+        WINDOWPLACEMENT wp = { .length = sizeof(wp), };
+        GetWindowPlacement(sys->hparent ? sys->hparent : sys->hwnd, &wp);
+        sys->desktop_save.win = wp.rcNormalPosition;
+    }
+
+    Direct3D11Close(vd);
+    EventThreadStop(sys->event);
+
+    vlc_mutex_lock(&sys->lock);
+    sys->use_desktop = sys->desktop_requested;
+    sys->ch_desktop = false;
+    vlc_mutex_unlock(&sys->lock);
+
+    event_cfg_t cfg;
+    memset(&cfg, 0, sizeof(cfg));
+    cfg.use_desktop = sys->use_desktop;
+    if (!sys->use_desktop) {
+        cfg.x      = sys->desktop_save.win.left;
+        cfg.y      = sys->desktop_save.win.top;
+        cfg.width  = sys->desktop_save.win.right  - sys->desktop_save.win.left;
+        cfg.height = sys->desktop_save.win.bottom - sys->desktop_save.win.top;
+    }
+
+    event_hwnd_t hwnd;
+    if (EventThreadStart(sys->event, &hwnd, &cfg)) {
+        msg_Err(vd, "Failed to restart event thread");
+        return VLC_EGENERIC;
+    }
+    sys->parent_window = hwnd.parent_window;
+    sys->hparent       = hwnd.hparent;
+    sys->hwnd          = hwnd.hwnd;
+    sys->hvideownd     = hwnd.hvideownd;
+    sys->hfswnd        = hwnd.hfswnd;
+    SetRectEmpty(&sys->rect_parent);
+
+    video_format_t fmt;
+    if (Direct3D11Open(vd, &fmt)) {
+        CommonClean(vd);
+        msg_Err(vd, "Failed to reopen device");
+        return VLC_EGENERIC;
+    }
+    vd->fmt = fmt;
+    sys->is_first_display = true;
+
+    if (sys->use_desktop) {
+        if (sys->desktop_save.is_fullscreen)
+            vout_display_SendEventFullscreen(vd, false);
+        if (sys->desktop_save.is_on_top)
+            vout_display_SendWindowState(vd, VOUT_WINDOW_STATE_NORMAL);
+    } else {
+        if (sys->desktop_save.is_fullscreen)
+            vout_display_SendEventFullscreen(vd, true);
+        if (sys->desktop_save.is_on_top)
+            vout_display_SendWindowState(vd, VOUT_WINDOW_STATE_ABOVE);
+    }
+    return VLC_SUCCESS;
+}
+
+static int Control(vout_display_t *vd, int query, va_list args)
+{
+    vout_display_sys_t *sys = vd->sys;
+
+    switch (query) {
+    case VOUT_DISPLAY_RESET_PICTURES:
+        if (sys->reset_device) {
+            if (ControlResetDevice(vd)) {
+                msg_Err(vd, "Failed to reset device");
+                return VLC_EGENERIC;
+            }
+            sys->reset_device = false;
+        } else if(sys->reopen_device) {
+            if (ControlReopenDevice(vd)) {
+                msg_Err(vd, "Failed to reopen device");
+                return VLC_EGENERIC;
+            }
+            sys->reopen_device = false;
+        }
+        return VLC_SUCCESS;
+    default:
+        return CommonControl(vd, query, args);
+    }
+}
+static void Manage (vout_display_t *vd)
+{
+    vout_display_sys_t *sys = vd->sys;
+
+    CommonManage(vd);
+
+    vlc_mutex_lock(&sys->lock);
+    const bool ch_desktop = sys->ch_desktop;
+    sys->ch_desktop = false;
+    vlc_mutex_unlock(&sys->lock);
+
+    if (ch_desktop) {
+        sys->reopen_device = true;
+        vout_display_SendEventPicturesInvalid(vd);
+    }
+
+    if (sys->changes & DX_POSITION_CHANGE) {
+
+        sys->clear_scene = true;
+        sys->changes &= ~DX_POSITION_CHANGE;
+    }
+}
+
+static HINSTANCE Direct3D9LoadShaderLibrary(void)
+{
+    HINSTANCE instance = NULL;
+    //d3dx11_46 is the latest on arm
+    for (int i = 46; i > 23; --i) {
+        TCHAR filename[16];
+        _sntprintf(filename, 16, TEXT("D3dx11_%d.dll"), i);
+        instance = LoadLibrary(filename);
+        if (instance)
+            break;
+    }
+    return instance;
+}
+
+static int Direct3D11Create(vout_display_t *vd)
+{
+    vout_display_sys_t *sys = vd->sys;
+
+    sys->hd3d11_dll = LoadLibrary(TEXT("D3D11.DLL"));
+    if (!sys->hd3d11_dll) {
+        msg_Warn(vd, "cannot load d3d11.dll, aborting");
+        return VLC_EGENERIC;
+    }
+    // TODO : load functions here
+    sys->hd3d11x_dll = Direct3D9LoadShaderLibrary();
+    if (!sys->hd3d11x_dll)
+        msg_Warn(vd, "cannot load Direct3D11 Shader Library; HLSL pixel shading will be disabled.");
+
+    return VLC_SUCCESS;
+}
+
+
+static void Direct3D11Destroy(vout_display_t *vd)
+{
+    vout_display_sys_t *sys = vd->sys;
+
+    // TODO : add release of d3d11 objects here
+    if (sys->hd3d11_dll)
+        FreeLibrary(sys->hd3d11_dll);
+    if (sys->hd3d11x_dll)
+        FreeLibrary(sys->hd3d11x_dll);
+
+    sys->hd3d11_dll = NULL;
+    sys->hd3d11x_dll = NULL;
+}
+
+static int Direct3D11Open(vout_display_t *vd, video_format_t *fmt)
+{
+    msg_Dbg(vd, "Direct3D11 device adapter successfully initialized");
+    return VLC_SUCCESS;
+}
+
+static void Direct3D11Close(vout_display_t *vd)
+{
+    msg_Dbg(vd, "Direct3D11 device adapter closed");
+}
+
+static int Direct3D11Reset(vout_display_t *vd)
+{
+    msg_Dbg(vd, "Direct3D11 device adapter reset");
+    return VLC_SUCCESS;
+}
+
+// TODO : d3d11 list supported formats by feature level 
+
+static int DesktopCallback(vlc_object_t *object, char const *psz_cmd,
+                            vlc_value_t oldval, vlc_value_t newval,
+                            void *p_data)
+{
+    vout_display_t *vd = (vout_display_t *)object;
+    vout_display_sys_t *sys = vd->sys;
+    VLC_UNUSED(psz_cmd);
+    VLC_UNUSED(oldval);
+    VLC_UNUSED(p_data);
+
+    vlc_mutex_lock(&sys->lock);
+    const bool ch_desktop = !sys->desktop_requested != !newval.b_bool;
+    sys->ch_desktop |= ch_desktop;
+    sys->desktop_requested = newval.b_bool;
+    vlc_mutex_unlock(&sys->lock);
+    return VLC_SUCCESS;
+}
diff --git a/modules/video_output/msw/events.c b/modules/video_output/msw/events.c
index b65ccb1..953d214 100644
--- a/modules/video_output/msw/events.c
+++ b/modules/video_output/msw/events.c
@@ -620,7 +620,7 @@ static void MouseReleased( event_thread_t *p_event, unsigned button )
     vout_display_SendEventMouseReleased( p_event->vd, button );
 }
 
-#ifdef MODULE_NAME_IS_direct3d
+#if defined(MODULE_NAME_IS_direct3d9) || defined(MODULE_NAME_IS_direct3d11)
 static int CALLBACK
 enumWindowsProc(HWND hwnd, LPARAM lParam)
 {
-- 
2.2.1

